import numpy as np
import os
from collections import namedtuple
import math
from scipy.interpolate import interp1d
from netCDF4 import Dataset
import scipy.io as sio


# Equatorial radius (6,378.1370 km)
# Polar radius (6,356.7523 km)
# The International Union of Geodesy and Geophysics 
# (IUGG) defines the mean radius (denoted R1) to be                        
# 6,371.009 km
R1_IUGG = 6371009

# Degree to radian conversion factor
deg2rad = np.pi / 180.


def center_point(lat, lon):
    """ centerpoint of lat/lon bounds """
    if isinstance(lat, (list, tuple, np.ndarray)):
        return (min(lat) + max(lat)) / 2, (min(lon) + max(lon)) / 2
    
    return lat, lon


def DLDL_over_DXDY(lat, lat_deriv_order, lon_deriv_order):
    """ Compute factor for transforming partial derivates in 
        lat-lon to partial derivates in x-y.

        Args: 
            lat: float or array
                Latitude of the positions(s) where the derivatives are to be evaluated
            lat_deriv_order: int
                Order of latitude-derivative
            lon_deriv_order: int
                Order of longitude-derivative

        Returns:
            ratio: float or array
                Factor for transforming partial derivates in lat-lon to partial derivates in x-y
    """
    R = R1_IUGG
    R2 = R * np.cos(lat * deg2rad)

    m = lat_deriv_order
    n = lon_deriv_order

    if m + n == 0:
        return 1

    ratio = 1
    
    if m > 0:
        ratio *= np.power(1./R, m)

    if n > 0:
        ratio *= np.power(1./R2, n)

    return ratio


def LLtoXY(lat, lon, lat_ref=0, lon_ref=0, rot=0, grid=False, z=None, squeeze=True):
    """ Transform lat-lon coordinates to xy position coordinates.

        By default, the origin of the xy coordinate system is 
        set to 0 deg latitude and 0 deg longitude. 
        
        By default, the x-axis is aligned with the longitude axis 
        (west to east) and the y-axis is aligned with the latitude
        axis (south to north).

        Args: 
            lat: float or numpy array
                latitude coordinate of a the location(s) of interest in degrees
            lon: float or numpy array
                longitude coordinate of a the location(s) of interest in degrees
            lat_ref: float
                latitude reference coordinate in degrees
            lon_ref: float
                longitude reference coordinate in degrees
            rot: float
                Rotation angle in degrees for the xy coordinate system 
                (clockwise rotation).
            squeeze: bool
                If output array has length 1, return float instead of array

        Returns:
            x: float or numpy array
                x position coordinates in meters
            y: float or numpy array
                y positions coordinates in meters
    """
    global R1_IUGG, deg2rad

    if grid:
        if z is None:
            lat, lon = np.meshgrid(lat, lon)
        else:
            lat, lon, z = np.meshgrid(lat, lon, z)

    else:
        if np.ndim(lat) == 0: lat = np.array([lat])
        if np.ndim(lon) == 0: lon = np.array([lon])
        if isinstance(lat, list): lat = np.array(lat)
        if isinstance(lon, list): lon = np.array(lon)

        assert lat.shape[0] == lon.shape[0], 'lat and lon must have same length'

    R = R1_IUGG
    R2 = R * np.cos(lat_ref * deg2rad)

    x = (lon - lon_ref) * deg2rad * R2
    y = (lat - lat_ref) * deg2rad * R

    if rot != 0:
        s = np.sin(rot * deg2rad)
        c = np.cos(rot * deg2rad)
        rotmat = np.array([[c, -s], [s, c]]) 
        xy = np.array([x, y])
        xy = np.swapaxes(xy, 0,1)
        xy = rotmat.dot(xy)
        x = xy[:,0]
        y = xy[:,1]

    if len(x) == 1 and squeeze:
        x = float(x)
        y = float(y)

    if z is None:
        return x, y
    else:
        return x, y, z

def XYtoLL(x, y, lat_ref=0, lon_ref=0, rot=0, grid=False, z=None, squeeze=True):
    """ Transform xy position coordinates to lat-lon coordinates.

        By default, the origin of the xy coordinate system is 
        set to 0 deg latitude and 0 deg longitude. 
        
        By default, the x-axis is aligned with the longitude axis 
        (west to east) and the y-axis is aligned with the latitude
        axis (south to north).

        Args: 
            x: float or numpy array
                x coordinate of a the location(s) of interest in meters
            y: float or numpy array
                y coordinate of a the location(s) of interest in meters
            lat_ref: float
                latitude reference coordinate in degrees
            lon_ref: float
                longitude reference coordinate in degrees
            rot: float
                Rotation angle in degrees for the xy coordinate system 
                (clockwise rotation).
            squeeze: bool
                If output array has length 1, return float instead of array

        Returns:
            lat: float or numpy array
                latitude coordinates in degrees
            lon: float or numpy array
                longitude coordinates in degrees
    """
    global R1_IUGG, deg2rad

    if grid:
        if z is None:
            x, y = np.meshgrid(x, y)
        else:
            x, y, z = np.meshgrid(x, y, z)

    else:
        if np.ndim(x) == 0: x = np.array([x])
        if np.ndim(y) == 0: y = np.array([y])
        if isinstance(x, list): x = np.array(x)
        if isinstance(y, list): y = np.array(y)

        assert x.shape[0] == y.shape[0], 'x and y must have same length'

    R = R1_IUGG
    R2 = R * np.cos(lat_ref * deg2rad)

    if rot != 0:
        s = np.sin(rot * deg2rad)
        c = np.cos(rot * deg2rad)
        rotmat = np.array([[c, s], [-s, c]]) 
        xy = np.array([x, y])
        xy = np.swapaxes(xy, 0,1)
        xy = rotmat.dot(xy)
        x = xy[:,0]
        y = xy[:,1]

    lon = lon_ref + x / deg2rad / R2
    lat = lat_ref + y / deg2rad / R

    if np.ndim(lat) == 1 and len(lat) == 1 and squeeze:
        lat = float(lat)

    if np.ndim(lon) == 1 and len(lon) == 1 and squeeze:
        lon = float(lon)

    if z is None:
        return lat, lon
    else:
        return lat, lon, z


def xdist(lon1, lon2, lat):
    R2 = R1_IUGG * np.cos(lat * deg2rad)
    d = np.abs(lon2 - lon1) * deg2rad * R2
    return d


def ydist(lat1, lat2):
    d = np.abs(lat2 - lat1) * deg2rad * R1_IUGG 
    return d


def torad(lat, lon):
    """ Convert latitute and longitude values from degrees to radians.

        The method expects the latitude to be in the range (-90,90) and
        the longitude to be in the range (-180,180).

        The output latitude is in the range (0,pi) and the output 
        longitude is in the range (-pi,pi).

        Args: 
            lat: float or array
                latitude(s) in degrees from -90 to +90.
            lon: float or array
                longitude(s) in degrees from -180 to +180.

        Returns:
            lat_rad: float or array
                latitude(s) in radians from 0 to pi.
            lon_rad: float or array
                longitude(s) in radians from -pi to pi.
    """
    lat_rad = (lat + 90) * deg2rad
    lon_rad = lon * deg2rad
    return lat_rad, lon_rad


def get_slices(distance, num_slices=1, bins=100, angle=0):
    """ Generate x,y coordinates for equally spaced radial slices 
        originating from (0,0).

        Args:
            distance: float
                Length of the radial slice.
            num_slices: int
                Number of slices
            bins: int
                Number of points per slice
            angle: float
                Angle of the first slice relative to the x-axis.

        Returns:
            x,y: list of numpy arrays
                x,y coordinate arrays for each slice
    """
    x,y = list(), list()

    # distance array
    dr = distance / float(bins)
    r = np.arange(bins, dtype=np.float)
    r *= dr
    r += 0.5 * dr

    # loop over angles
    a = angle
    da = 360. / float(num_slices)
    for _ in range(num_slices):
        x.append(r * np.cos(a * np.pi / 180.))
        y.append(r * np.sin(a * np.pi / 180.))

    if num_slices == 1:
        x = x[0]
        y = y[0]
    
    return x, y


def interp_grid_1d(y, x=None, num_pts=math.inf, rel_err=None, method='linear'):
    """ Determine the optimal interpolation grid for the 
        function y(x). 
        
        The grid will in general not be uniform, as the 
        grid points will be more densily clustered in 
        regions where y(x) is changing more rapidly. 
        
        Args:
            y: 1d numpy array
                y values
            x: 1d numpy array
                x values. If none are specified, they are 
                assumed to be 0,1,2,...
            num_pts: int
                Number of grid points. If rel_err is specified, 
                num_pts becomes the maximum possible number of 
                grid points.
            rel_err: float
                Maximum deviation between the interpolation and 
                y, relative to the range of values spanned by y.
            method: str
                Interpolation method

        Returns:
            a: 1d numpy array
                Indices of the grid points
            e: float
                Maximum relative error
    """
    n = len(y)
    norm = np.max(y) - np.min(y)

    if x is None:
        x = np.arange(n)
        
    if num_pts == math.inf and rel_err is None:
        num_pts = 101

    a = np.array([0, n-1])
    num = len(a)

    while num < num_pts:
        
        f = interp1d(x=x[a], y=y[a], kind=method)
        
        dev = np.abs(y - f(x))
        
        a0 = np.argmax(dev)
        dev0 = dev[a0]
        
        e = dev0 / norm
        
        if rel_err is not None and e < rel_err:
            break
        else:
            a = np.append(a, a0)

        num = len(a)

    a = np.sort(a)
    return a, e


def load_data_from_file(path, val_name='bathy', lat_name='lat', lon_name='lon', lon_axis=1,\
    south=-90, north=90, west=-180, east=180):
    """ Load geospatial data from a single file. 

        Currently supported formats are NetCDF (*.nc) and MatLab (*.mat).

        The data can be cropped by speciyfing south/north/west/east 
        boundaries.

        Args: 
            path: str
                File path
            val_name: str
                Name of variable/field containing the data values
            lat_name: str
                Name of variable/field containing the latitude values
            lon_name: str
                Name of variable/field containing the longitude values
            lon_axis: int
                Specify if the longitude dimension is the second (1, default) 
                or first (0) axis.
            south: float
                Southern boundary of the region of interest.
            north: float
                Northern boundary of the region of interest.
            west: float
                Western boundary of the region of interest.
            east: float
                Eastern boundary of the region of interest.

        Returns:
            val: 1d or 2d numpy array
                Data values
            lat: numpy array
                Latitude values
            lon: numpy array
                Longitude values
    """
    # detect format
    ext = path[path.rfind('.'):]

    # load data
    if ext == '.nc': # NetCDF
        d = Dataset(path)
        val = np.array(d[val_name])
        lat = np.array(d[lat_name])
        lon = np.array(d[lon_name])

    elif ext == '.mat': # MatLab
        d = sio.loadmat(path)
        val = np.array(d[val_name])
        lat = np.squeeze(np.array(d[lat_name]))
        lon = np.squeeze(np.array(d[lon_name]))

    else:
        print('Unknown file format *{0}'.format(ext))
        exit(1)

    # crop the region of interest
    grid = (np.ndim(val) == 2)
    indices, lat, lon = crop(lat, lon, south, north, west, east, grid=grid)
    val = val[indices]

    # ensure that lat and lon are strictly increasing
    if np.all(np.diff(lat) < 0):
        lat = np.flip(lat, axis=0)
        val = np.flip(val, axis=0)
    if np.all(np.diff(lon) < 0):
        lon = np.flip(lon, axis=0)
        val = np.flip(val, axis=1)

    # flip axes, if necessary
    if lon_axis == 0:
        val = np.swapaxes(val, 0, 1)

    # if axis size are inconsistent, try swapping
    if val.shape[0] != lat.shape[0]:
        val = np.swapaxes(val, 0, 1)

    return val, lat, lon


def crop(lat, lon, south, north, west, east, grid=False):
    """ Select rectangular region bounded by the geographical coordinates 
        latlon_SW to the south-west and latlon_NE to the north-east.

        If grid is False, lat and lon must have the same length.

        Args: 
            lat: 1d or 2d numpy array
                Latitude values
            lon: 1d or 2d numpy array
                Longitude values
            south: float
                Southern boundary of the region of interest.
            north: float
                Northern boundary of the region of interest.
            west: float
                Western boundary of the region of interest.
            east: float
                Eastern boundary of the region of interest.
            grid: bool
                Specify how to combine elements of lat and lon.

        Returns:
            ind: numpy array
                Selected indices. 1d if grid is False, 2d if grid is True.
            lat: numpy array
                Latitude values
            lon: numpy array
                Longitude values
    """
    ind_lat = np.argwhere((lat >= south) & (lat <= north))
    ind_lat = np.squeeze(ind_lat)

    ind_lon = np.argwhere((lon >= west) & (lon <= east))
    ind_lon = np.squeeze(ind_lon)

    if np.ndim(ind_lat) == 2:
        latc = ind_lat[:,0] + 1j * ind_lat[:,1]
        lonc = ind_lon[:,0] + 1j * ind_lon[:,1]
        x = np.intersect1d(latc, lonc)
        xr = np.real(x)
        xi = np.imag(x)
        xr = xr.astype(int)
        xi = xi.astype(int)
        ind_lat = np.arange(np.min(xr), np.max(xr)+1)
        ind_lon = np.arange(np.min(xi), np.max(xi)+1)
        lat = lat[:,0]
        lon = lon[0,:]
        ind = np.ix_(ind_lat, ind_lon)       

    if grid:
        ind = np.ix_(ind_lat, ind_lon)       
    else:
        ind = np.intersect1d(ind_lat, ind_lon)
        ind_lat = ind
        ind_lon = ind

    lat = lat[ind_lat]
    lon = lon[ind_lon]

    return ind, lat, lon


def create_boolean_array(n, step=1):
    arr = np.zeros(n, dtype=bool)
    if step == math.inf: arr[0] = True
    else: arr[::step] = True
    return arr


def toarray(x):
    if isinstance(x, float) or isinstance(x, int): x = [x]    
    if isinstance(x, list): x = np.array(x)    
    return x



""" Unit tests for the the 'geospatial.ocean' module in the 'kadlu' package

    Authors: Oliver Kirsebom
    contact: oliver.kirsebom@dal.ca
    Organization: MERIDIAN-Intitute for Big Data Analytics
    Team: Acoustic data Analytics, Dalhousie University
    Project: packages/kadlu
             Project goal: Tools for underwater soundscape modeling
     
    License:

"""

import os
import logging
from datetime import datetime

import pytest
import numpy as np
import pandas as pd
from kadlu.geospatial.ocean import Ocean
from kadlu.geospatial.data_sources.source_map import default_val
from kadlu import hycom


bounds = dict(
        start=datetime(2015, 1, 9), end=datetime(2015, 1, 9, 3),
        south=44,                   west=-64.5, 
        north=46,                   east=-62.5, 
        top=0,                      bottom=5000
    )

test_lat, test_lon, test_depth = bounds['south'], bounds['west'], bounds['top']

def test_null_ocean():
    """ Test that ocean is initialized with all variables set to 
        null (0) when default=False"""
    #o = Ocean(default=False, cache=False)
    
    # changed to make ocean null by default
    o = Ocean(**bounds)
    
    assert o.bathy(test_lat, test_lon) == 0
    assert o.temp(test_lat, test_lon, test_depth) == 0
    assert o.salinity(test_lat, test_lon, test_depth) == 0
    assert o.wavedir(test_lat, test_lon) == 0
    assert o.waveheight(test_lat, test_lon) == 0
    assert o.waveperiod(test_lat, test_lon) == 0
    assert o.wind_uv(test_lat, test_lon) == 0
    assert o.origin == (45, -63.5)
    assert o.boundaries == bounds

def test_uniform_bathy():
    """ Test that ocean can be initialized with uniform bathymetry"""
    #o = Ocean(default=False, cache=False, load_bathymetry=-500.5)
    o = Ocean(load_bathymetry=500.5, **bounds)

    assert o.bathy(test_lat, test_lon) == 500.5
    assert o.temp(test_lat, test_lon, test_depth) == 0

def test_interp_uniform_temp():
    """ Test that we can interpolate a uniform ocean temperature 
        on any set of coordinates"""
    #o = Ocean(default=False, cache=False, load_temp=16.1)
    o = Ocean(load_temp=16.1, **bounds)
    assert o.temp(lat=41.2, lon=-66.0, depth=-33.0) == 16.1
    #assert o.temp_xy(x=1, y=2.2, z=-3.0) == 16.1
    #assert np.all(o.temp_xy(x=[5,20], y=[0,10], z=[-300,-400]) == [16.1, 16.1])

def test_uniform_bathy_deriv():
    """ Test that uniform bathy has derivative zero"""
    #o = Ocean(default=False, cache=False, load_bathymetry=-500.5)
    o = Ocean(load_bathymetry=-500.5, **bounds)
    assert o.bathy_deriv(lat=1,lon=17,axis='lon') == 0

def test_chs_bathy():
    """ Test that ocean can be initialized with bathymetry data 
        from a CHS file with automatic fetching enabled"""
    bound_args = bounds.copy()
    bound_args['south'], bound_args['west'], bound_args['north'], bound_args['east'] = 43.1, -59.8, 43.8, -59.2
    o = Ocean(fetch=8, load_bathymetry='chs', **bound_args)
    test_lat = [43.4, 43.5]
    test_lon = [-59.6, -59.5]
    bathy = o.bathy(test_lat, test_lon)
    assert len(bathy) > 0 #check that some data was retrieved
    assert  43.1 <= np.min(test_lat) and np.max(test_lat) <=  43.8 #check that lats are within limits
    assert -59.8 <= np.min(test_lon) and np.max(test_lon) <= -59.2 #check that lons are within limits
    # check that all nodes have meaningful bathymetry values
    assert np.all(bathy < 10000)
    assert np.all(bathy > -15000)
    assert o.interps['bathy'].origin == o.origin

def test_interp_chs_bathy():
    """ Test that we can interpolate bathymetry data 
        obtained from a CHS file"""
    o = Ocean(load_bathymetry='chs', 
            south=43.1, west=-59.8, north=43.8, east=-59.2, 
            top=0, bottom=0, start=default_val['start'], end=default_val['end'])
    b = o.bathy(lat=1, lon=2)

    assert isinstance(b, float)
    b = o.bathy(lat=[43.2,43.7], lon=[-59.3, -59.4])
    assert len(b) == 2

def test_interp_hycom_temp():
    """ Test that ocean can be initialized with temperature data 
        from HYCOM with automatic fetching enabled and using 
        start/end args.
    """
    kwargs = dict(
            south=43.1, west=-59.8, 
            north=43.8, east=-59.2,
            top=-100, bottom=3000,
            start=datetime(2015,1,1),
            end=datetime(2015,1,2)
        )
    o = Ocean(#fetch=True, #cache=False,
        load_temp='hycom', 
        load_bathymetry='chs',
        **kwargs
        )
    lats = [43.4, 43.5]
    lons = [-59.6, -59.5]
    depths = [200, 300]
    #(temp,lats,lons,depths) = o.temp()
    temp = o.temp(lats, lons, depths)
    assert  43.1 <= np.min(lats) and np.max(lats) <=  43.8 #check that lats are within limits
    assert -59.8 <= np.min(lons) and np.max(lons) <= -59.2 #check that lons are within limits
    assert 3000 >= np.max(depths) and np.min(depths) >= -100 #check that depths are within limits
    assert len(temp) > 0   #check that some data was retrieved
    assert len(depths) > 1 #check that more than one depth was fetched
    assert np.all(np.abs(temp - temp.astype(int) > 1e-12)) # check that temperatures are not integers
    # interpolate temperature at seafloor
    seafloor_depth = o.bathy(lats, lons)
    temp = o.temp(lats, lons, seafloor_depth)
    assert np.all(np.logical_and(temp > -5, temp < 105)) # check sensible temperature at bottom

    # load hycom temperature data
    temp, lat, lon, epoch, depth = hycom().load_temp(**kwargs)
    # select a location close to the center of the region
    df = pd.DataFrame({'temp': temp, 'lat': lat, 'lon': lon, 'epoch': epoch, 'depth': depth})
    lat_close = df.lat.values[np.argmin(np.abs(df.lat.values - 43.45))]
    lon_close = df.lon.values[np.argmin(np.abs(df.lon.values - 59.5))]
    df = df[(df.lat == lat_close) & (df.lon == lon_close)]
    # query temps from ocean interpolator    
    depths = np.unique(df.depth.values)
    temp_ocean = o.temp(lat_close, lon_close, depths, grid=True)
    df = df.set_index(['depth','epoch'])
    df = df.groupby(level=[0]).mean()
    # check that interpolated temps and original temps agree
    assert np.all(temp_ocean == df.temp.values)

def test_array_bathy():
    """ Test that ocean can be initialized with bathymetry data 
        from arrays"""
    bathy = np.array([[-100., -200.],
                      [-100., -200.]])
    lats = np.array([44.5, 44.7])
    lons = np.array([-60.1, -59.5])
    # note that fetching does nothing when supplying raw array data
    o = Ocean(load_bathymetry=(bathy, lats, lons), fetch=True, **bounds)
    la = lats
    lo = lons
    b = o.bathy(lat=lats, lon=lons)
    assert np.all(b == bathy)
    assert np.all(la == lats)
    assert np.all(lo == lons)
    res = o.bathy(lat=44.5, lon=-60.1)
    assert res == -100
    res = o.bathy(lat=44.5, lon=-59.8)
    assert pytest.approx(res == -150., abs=1e-6)

def test_small_full_ocean():
    """ test that the ocean can be initialized for a very small region """

    bounds = dict(
            start=datetime(2015, 1, 9), end=datetime(2015, 1, 9, 3),
            south=44.2,                 west=-64.4, 
            north=44.21,                east=-64.39, 
            top=0,                      bottom=1
        )
    try:
        o = Ocean(load_bathymetry='chs', load_temp='hycom', load_salinity='hycom', 
                load_wavedir='era5', load_waveheight='wwiii', load_waveperiod='era5', 
                load_wind_uv='wwiii', **bounds)
    except AssertionError as err:
        # this is intended behaviour
        logging.info('CAUGHT EXCEPTION: ' + str(err))
        pass 
    except Exception as err:
        raise err


def test_wind_water_uv():
        o = Ocean(load_water_u='hycom', load_water_v='hycom', load_water_uv='hycom', 
                load_wind_u='era5', load_wind_v='era5', load_wind_uv='era5', 
                **bounds)
    




"""
    Utilities for data fetching and loading.
"""

import os
import sys
import json
import pickle
import sqlite3
import logging
import warnings
import configparser
from os import path
from os.path import dirname
from hashlib import md5
from functools import reduce
from datetime import datetime, timedelta
from contextlib import contextmanager, redirect_stdout, redirect_stderr

import numpy as np


LOGLEVEL = os.environ.get('LOGLEVEL', 'INFO')
logging.basicConfig(format='%(asctime)s  %(message)s', level=LOGLEVEL, datefmt='%Y-%m-%d %I:%M:%S')


# database tables for data fetching and loading
chs_table    = 'chs_bathy'
#hycom_tables = ['salinity', 'water_temp', 'water_u', 'water_v']
wwiii_tables = ['hs', 'dp', 'tp', 'windU', 'windV']
era5_tables  = [
        'significant_height_of_combined_wind_waves_and_swell',
        'mean_wave_direction',
        'mean_wave_period',
        'u_component_of_wind',
        'v_component_of_wind'
    ]


cfg = configparser.ConfigParser()       # read .ini into dictionary object
cfgfile = os.path.join(dirname(dirname(dirname(dirname(__file__)))), "config.ini")
cfg.read(cfgfile)


def storage_cfg(setdir=None):
    """ return filepath containing storage configuration string

        first checks the config.ini file in kadlu root folder, then
        defaults to kadlu/storage
    """

    def default_storage(msg):
        """ helper function for storage_cfg() """
        storage_location = os.path.join(os.path.expanduser('~'), f'kadlu_data{os.path.sep}')
        if not os.path.isdir(storage_location): os.mkdir(storage_location)
        warnings.warn(f'{msg} storage location will be set to {storage_location}')
        return storage_location

    if 'storage' not in cfg.sections():
        cfg.add_section('storage')

    if setdir is not None:
        assert os.path.isdir(setdir), 'directory does not exist'
        cfg.set('storage', 'storage_location', setdir)
        with open(cfgfile, 'w') as f:
            cfg.write(f)

    try:
        storage_location = cfg['storage']['storage_location']
    except KeyError:                        # missing config.ini file
        return default_storage('storage location not configured.')

    if storage_location == '':              # null value in config.ini
        return default_storage('null value in config.')

    if not path.isdir(storage_location):    # verify the location exists
        return default_storage('storage location does not exist.')

    if storage_location[-1] != os.path.sep: 
        return storage_location + os.path.sep

    return storage_location


def database_cfg():
    """ configure and connect to sqlite database

        time is stored as an integer in the database, where each value
        is epoch hours since 2000-01-01 00:00

        returns:
            conn:   
                database connection object
            db:
                connection cursor object
    """
    conn = sqlite3.connect(storage_cfg() + 'geospatial.db')
    db = conn.cursor()

    # bathymetry table (CHS)
    db.execute(f'CREATE TABLE IF NOT EXISTS {chs_table} ' 
                '(   val     REAL    NOT NULL, ' 
                '    lat     REAL    NOT NULL, ' 
                '    lon     REAL    NOT NULL, ' 
                '    source  TEXT    NOT NULL) ') 
    db.execute(f'CREATE UNIQUE INDEX IF NOT EXISTS '
               f'idx_{chs_table} on {chs_table}(lon, lat, val, source)')


    # wave data tables
    for var in era5_tables + wwiii_tables:
        db.execute(f'CREATE TABLE IF NOT EXISTS {var}'
                    '( val     REAL    NOT NULL, ' 
                    '  lat     REAL    NOT NULL, ' 
                    '  lon     REAL    NOT NULL, ' 
                    '  time    INT     NOT NULL, ' 
                    '  source  TEXT    NOT NULL) ') 
        db.execute(f'CREATE UNIQUE INDEX IF NOT EXISTS '
                   f'idx_{var} on {var}(time, lon, lat, val, source)')

    return conn, db


def bin_db():
    """ database for storing serialized objects in memory 
        
        code was previously used for caching results quickly but
        is currently not in use
    """
    conn = sqlite3.connect('file::memory:?cache=shared', uri=True)
    db = conn.cursor()
    db.execute('CREATE TABLE IF NOT EXISTS bin'
                '(  hash    INT  NOT NULL,  '
                '   bytes   BLOB NOT NULL  )' )
    db.execute('CREATE UNIQUE INDEX IF NOT EXISTS' 
               'idx_bin on bin(hash)')

    raise ResourceWarning('fcn not used')


def hash_key(kwargs, seed, 
        keys=('south','west', 'north', 'east', 'top', 'bottom', 'start', 'end')):
    """ compute unique hash and convert to 8-byte int as serialization key """

    # copy boundary arguments to json string
    qry = kwargs.copy()
    dim = kwargs.keys()
    for dimension in dim:
        if dimension not in keys: del qry[dimension]
    string = seed + json.dumps(qry, sort_keys=True, default=str)

    # interpret json hash as base-16 integer and bitshift to fit SQL MAXINT
    key = int(md5(string.encode('utf-8')).hexdigest(), base=16) >> 80

    return key


def insert_hash(kwargs={}, seed='', obj=None):
    """ create hash index in database to record query history.
        this is used for mapping the coverage of fetched data,
        optionally include an object to be serialized and cached
    """
    qry = kwargs.copy()
    if 'lock' in qry.keys(): del qry['lock']
    key = hash_key(qry, seed)
    #conn, db = database_cfg()
    conn = sqlite3.connect(storage_cfg() + 'checksums.db')
    db = conn.cursor()
    db.execute('CREATE TABLE IF NOT EXISTS fetch_map'
                '(  hash    INT  NOT NULL, '
                '   bytes   BLOB         ) ' )
    db.execute(f'CREATE UNIQUE INDEX IF NOT EXISTS '
                 f'idx_fetched on fetch_map(hash)')
    db.execute('INSERT OR IGNORE INTO fetch_map VALUES (?,?)',
               (key, pickle.dumps(obj)))
    conn.commit()
    return


def serialized(kwargs={}, seed=''):
    """ returns true if fetch query hash exists in database else False """
    key = hash_key(kwargs, seed)
    if 'lock' in kwargs.keys(): kwargs['lock'].acquire()
    #conn, db = database_cfg()
    conn = sqlite3.connect(storage_cfg() + 'checksums.db')
    db = conn.cursor()
    db.execute('CREATE TABLE IF NOT EXISTS fetch_map'
                '(  hash    INT  NOT NULL, '
                '   bytes   BLOB         ) ' )
    db.execute(f'CREATE UNIQUE INDEX IF NOT EXISTS '
                 f'idx_fetched on fetch_map(hash)')
    db.execute('SELECT * FROM fetch_map WHERE hash == ?', (key,))
    res = db.fetchone()
    if 'lock' in kwargs.keys(): kwargs['lock'].release()
    if res is None: return False
    if res[1] is not None: return res[1]
    return True


def dt_2_epoch(dt_arr):
    """ convert datetimes to epoch hours """
    t0 = datetime(2000, 1, 1, 0, 0, 0)
    delta = lambda dt: (dt - t0).total_seconds()/60/60
    if isinstance(dt_arr, (list, np.ndarray)): return list(map(int, map(delta, dt_arr)))
    elif isinstance(dt_arr, (datetime)): return int(delta(dt_arr))
    else: raise ValueError('input must be datetime or array of datetimes')


def epoch_2_dt(ep_arr):
    """ convert epoch hours to datetimes """
    t0 = datetime(2000, 1, 1)
    delta = lambda ep : t0 + timedelta(hours=ep)
    if isinstance(ep_arr, (list, np.ndarray)): return list(map(delta, ep_arr))
    elif isinstance(ep_arr, (float, int)): return delta(ep_arr)
    else: raise ValueError('input must be integer or array of integers')


def index(val, sorted_arr):
    """ converts value in coordinate array to grid index """
    if val > sorted_arr[-1]: return len(sorted_arr) - 1
    return np.nonzero(sorted_arr >= val)[0][0]


def flatten(cols, frames):
    """ dimensional reduction by taking average of time frames """
    # assert that frames are of equal size
    assert len(cols) == 5
    assert reduce(lambda a, b: (a==b)*a, frames[1:] - frames[:-1])

    # aggregate time frame splits and reduce them to average
    frames_ix = np.append([0], frames)
    split_num = range(len(frames_ix) -1)
    val_split = np.array([cols[0][frames_ix[f] : frames_ix[f+1]] for f in split_num])
    value_avg = (reduce(np.add, val_split) / len(val_split))
    _, y, x, _, z = cols[:, frames_ix[0] : frames_ix[1]]
    return value_avg, y, x, z


def reshape_2D(cols):
    return dict(values=cols[0], lats=cols[1], lons=cols[2])


def reshape_3D(cols):
    """ prepare loaded data for interpolation """
    if isinstance(cols[0], (float, int)):
        return dict(values=cols[0])

    # if data is 4D, take average of values at time frame intervals
    frames = np.append(np.nonzero(cols[3][1:] > cols[3][:-1])[0] + 1, len(cols[3]))
    if len(np.unique(frames)) > 1: vals, y, x, z = flatten(cols, frames) 
    else: vals, y, x, _, z  = cols
    rows = np.array((vals, y, x, z)).T

    # reshape row data to 3D array
    xgrid, ygrid, zgrid = np.unique(x), np.unique(y), np.unique(z)
    gridspace = np.full((len(ygrid), len(xgrid), len(zgrid)), fill_value=None, dtype=float)
    # this could potentially be optimized to avoid an index lookup cost
    for row in rows:
        x_ix = index(row[2], xgrid)
        y_ix = index(row[1], ygrid)
        z_ix = index(row[3], zgrid)
        gridspace[y_ix, x_ix, z_ix] = row[0]

    # remove nulls for interpolation:
    # fill missing depth values with last value in each column
    for xi in range(0, gridspace.shape[0]):
        for yi in range(0, gridspace.shape[1]):
            col = gridspace[xi, yi]
            if sum(np.isnan(col)) > 0 and sum(np.isnan(col)) < len(col):
                col[np.isnan(col)] = col[~np.isnan(col)][-1]
                gridspace[xi, yi] = col

    # null depth columns are filled with the average value at each depth plane
    for zi in range(0, gridspace.shape[2]):
        gridspace[:,:,zi][np.isnan(gridspace[:,:,zi])] = np.average(gridspace[:,:,zi][~np.isnan(gridspace[:,:,zi])])

    return dict(values=gridspace, lats=ygrid, lons=xgrid, depths=zgrid)


def str_def(self, info, args):
    """ builds string definition for data source class objects """
    fcns = [fcn for fcn in dir(self) if callable(getattr(self, fcn)) 
            and 'load' in fcn and not fcn.startswith("__")]
    return (f'{info}\n\nfunction input arguments:\n\t{args}\n\nclass functions:\n\t'
            + '\n\t'.join(fcns) + '\n')

def era5_cfg(key=None, url=None):
    if 'cdsapi' not in cfg.sections():
        cfg.add_section('cdsapi')

    if key is not None:
        cfg.set('cdsapi', 'key', key)
        with open(cfgfile, 'w') as f:
            cfg.write(f)

    if url is not None:
        cfg.set('cdsapi', 'url', url)
        with open(cfgfile, 'w') as f:
            cfg.write(f)

    return 


def ll_2_regionstr(south, north, west, east, regions, default=[]):
    """ convert input bounds to region strings with seperating axis theorem """
    if west > east:  # recursive function call if query intersects antimeridian
        return np.union1d(ll_2_regionstr(south, north, west, +180, regions, default), 
                          ll_2_regionstr(south, north, -180, east, regions, default))

    query = Boundary(south, north, west, east)
    matching = [str(reg) for reg in regions if query.intersects(reg)]

    if len(matching) == 0: 
        warnings.warn(f"No regions matched for query. Defaulting to {default} ({len(default)} regions)")
        return default

    return np.unique(matching)


def fmt_coords(kwargs):
      return (
            f"{abs(kwargs['south']):.2f}°{'S' if kwargs['south'] <= 0 else 'N'},"
            f"{abs(kwargs['west']):.2f}°{'W' if kwargs['west'] <= 0 else 'E'}"
            f"{','+str(kwargs['top'])+'m' if 'top' in kwargs.keys() else ''}:"
            f"{abs(kwargs['north']):.2f}°{'S' if kwargs['north'] <= 0 else 'N'},"
            f"{abs(kwargs['east']):.2f}°{'W' if kwargs['east'] <= 0 else 'E'}"
            f"{','+str(kwargs['bottom'])+'m' if 'bottom' in kwargs.keys() else ''}"
        )



class Boundary():
    """ compute intersecting boundaries with separating axis theorem """
    def __init__(self, south, north, west, east, fetchvar=''):
        self.south, self.north, self.west, self.east, self.fetchvar = \
                south, north, west, east, fetchvar

    def __str__(self): return self.fetchvar

    def intersects(self, other):
        return not (self.east  < other.west or
                    self.west  > other.east or
                    self.north < other.south or
                    self.south > other.north)


@contextmanager
def dev_null():
    """ context manager to redirect output to /dev/null """
    with open(os.devnull, 'w') as null:
        try:
            with redirect_stderr(null) as err, redirect_stdout(null) as out: 
                yield (err, out)
        finally:
            sys.stdout = sys.__stdout__
            sys.stderr = sys.__stderr__


"""
class DataUtil():
    # user API for data utils 
    def epoch_2_dt(self, arr):              return epoch_2_dt(arr)
    def dt_2_epoch(self, arr):              return dt_2_epoch(arr)
    def index(self, sorted_arr):            return index(sorted_arr)
    def reshape_2D(self, cols):             return reshape_2D(cols)
    def reshape_3D(self, cols):             return reshape_3D(cols)
    def storage_cfg(self, setdir=None):     return storage_cfg(setdir)
    def database_cfg(self):                 return database_cfg()
    def era5_cfg(self, key=None, url=None): return era5_cfg(key, url)
"""





from setuptools import setup, find_packages

# create distribution and upload to pypi.org with:
#   $ python setup.py sdist bdist_wheel
#   $ twine upload dist/*

setup(name='kadlu',
        version='2.1.0',
        description="MERIDIAN Python package for ocean ambient noise modelling",
        url='https://gitlab.meridian.cs.dal.ca/public_projects/kadlu',
        author='Oliver Kirsebom, Matthew Smith',
        author_email='oliver.kirsebom@dal.ca, matthew.smith@dal.ca',
        license='GNU General Public License v3.0',
        packages=find_packages(),
        install_requires=[
            'numpy',
            'scipy',
            'pytest',
            'pyproj',
            'pygrib',
            'cdsapi',
            'matplotlib',
            'geos',     # needed for cartopy
            'proj',     # needed for cartopy
            'cartopy',
            'Pillow',
            'imageio',
            ],
        setup_requires=['pytest-runner',],
        tests_require=['pytest',],
        include_package_data=True,
        python_requires='>=3.8',
        #zip_safe=False
    )




import pytest
import logging
from datetime import datetime, timedelta
from kadlu.geospatial.data_sources import wwiii
from kadlu.geospatial.data_sources.wwiii import Wwiii, Boundary, wwiii_regions, wwiii_global


start   = datetime(2014, 2, 3, 0, 0, 0, 0)
end     = datetime(2014, 2, 3, 3, 0, 0, 0)

# gulf st lawrence
south, west = 47,   -64
north, east = 48,   -63


def test_wwiii_ll2regionstr():
    # gulf st lawrence
    south =  46
    north =  52
    west  = -70
    east  = -56
    regions = wwiii.ll_2_regionstr(south, north, west, east, wwiii_regions, wwiii_global)
    assert(len(regions) == 1)
    assert(regions[0] == 'at_4m')

    # bering sea
    # test area intersection across antimeridian 
    south, north = 46, 67
    west, east = 158, -156
    east=-156
    regions = wwiii.ll_2_regionstr(south, north, west, east, wwiii_regions, wwiii_global)
    assert(len(regions) == 3)
    assert('ak_4m' in regions)
    assert('ao_30m' in regions)
    assert('wc_4m' in regions)

    # global 
    globe = wwiii.ll_2_regionstr(-90, 90, -180, 180, wwiii_regions, wwiii_global)
    assert(len(globe) == 5)

"""
def test_wwiii_fetch_windwaveheight():
    if not Wwiii().fetch_windwaveheight(south=south, north=north, west=west, east=east, start=start, end=end):
        print('wwiii query was fetched already, skipping...')
    return 

def test_wwiii_fetch_wavedirection():
    if not Wwiii().fetch_wavedirection(south=south, north=north, west=west, east=east, start=start, end=end):
        print('wwiii query was fetched already, skipping...')
    return 
def test_wwiii_fetch_waveperiod():
    if not Wwiii().fetch_waveperiod(south=south, north=north, west=west, east=east, start=start, end=end):
        print('wwiii query was fetched already, skipping...')
    return 

def test_wwiii_fetch_windwaveheight():
    if not Wwiii().fetch_windwaveheight(south=south, north=north, west=west, east=east, start=start, end=end):
        print('wwiii query was fetched already, skipping...')
    return 
"""

def test_wwiii_load_windwaveheight():
    wave, lat, lon, time = Wwiii().load_windwaveheight(south=south, west=west, north=north, east=east, start=start, end=end)


"""
def test_wwiii_fetch_wind():
    if not Wwiii().fetch_wind_uv(south=south, north=north, west=west, east=east, start=start, end=end):
        print('wwiii query was fetched already, skipping...')
    return 
"""

def test_wwiii_load_wind():
    try:
        wave, lat, lon, time = Wwiii().load_wind_uv(south=south, west=west, north=north, east=east, start=start, end=end)
    except AssertionError as err:
        logging.info(f'CAUGHT EXCEPTION: {str(err)}')
    except Exception as err:
        raise err

"""
def test_wwiii_fetch_waveperiod():
    if not Wwiii().fetch_waveperiod(south=south, north=north, west=west, east=east, start=start, end=end):
        print('wwiii query was fetched already, skipping...')
"""

def test_wwiii_load_waveperiod():
    wave, lat, lon, time = Wwiii().load_waveperiod(south=south, west=west, north=north, east=east, start=start, end=end)
  




"""
    API for NONNA-100 Bathymetric Data from the Canadian Hydrographic Service (CHS)
     
    Metadata regarding the dataset can be found here:
        https://open.canada.ca/data/en/dataset/d3881c4c-650d-4070-bf9b-1e00aabf0a1d
"""

import os
import json
import logging
import requests
import warnings
from PIL import Image
from datetime import datetime

import numpy as np

import kadlu.geospatial.data_sources.fetch_handler
from kadlu.geospatial.data_sources.data_util        import          \
        database_cfg,                                               \
        storage_cfg,                                                \
        insert_hash,                                                \
        serialized,                                                 \
        fmt_coords,                                                 \
        chs_table,                                                  \
        str_def


conn, db = database_cfg()


def parse_sw_corner(path):
    """ return the southwest corner coordinates for a given bathymetry file """
    fname = os.path.basename(path)
    south = int(fname[4:8]) / 100
    west = -int(fname[9:14]) / 100
    assert west >= -180 and west <= 180, 'Invalid parsed longitude value'
    assert south >= -90 and south <= 90, 'Invalid parsed latitude value'
    return south, west


def fetch_chs(south, north, west, east, band_id=1):
    """ download bathymetric geotiffs, process them, and insert into db

        args:
            south, north: float
                ymin, ymax coordinate boundaries. range: -90, 90
            west, east: float
                xmin, xmax coordinate boundaries. range: -180, 180

        return: 
            True if new data was downloaded and processed, else False
    """

    # api call: get raster IDs within bounding box
    source = "https://gisp.dfo-mpo.gc.ca/arcgis/rest/services/FGP/CHS_NONNA_100/"
    spatialRel = "esriSpatialRelIntersects"
    spatialReference = "4326"  # WGS-84 spec
    geometry = json.dumps({"xmin":west, "ymin":south, "xmax":east, "ymax":north})
    url1 = f"{source}ImageServer/query?geometry={geometry}&returnIdsOnly=true&geometryType=esriGeometryEnvelope&spatialRel={spatialRel}&f=json&outFields=*&inSR={spatialReference}"
    req1 = requests.get(url1)
    assert(req1.status_code == 200)
    assert("error" not in json.loads(req1.text).keys())

    # api call: query for resource locations of rasters
    imgs = []
    rasterIds = json.loads(req1.text)['objectIds']
    assert(len(rasterIds) > 0)
    for chunk in range(0, int(len(rasterIds) / 20) + 1):  # max request size is 20 at a time
        rasterIdsCSV = ','.join([f"{x}" for x in rasterIds[chunk * 20:(chunk+1) * 20]])
        url2 = f"{source}ImageServer/download?geometry={geometry}&geometryType=esriGeometryPolygon&format=TIFF&f=json&rasterIds={rasterIdsCSV}"
        req2 = requests.get(url2)
        assert(req2.status_code == 200)
        jsondata = json.loads(req2.text)
        assert("error" not in jsondata.keys())
        imgs += [img for img in jsondata['rasterFiles'] if 'CA2' in img['id']]

    # api call: for each tiff image, download the associated rasters
    filepaths = []
    imgnum = 1
    for img in imgs:
        fname = img['id'].split('\\')[-1]
        fpath = f"{storage_cfg()}{fname}"
        filepaths.append(fpath)
        if os.path.isfile(fpath): 
            logging.info(f'CHS {fname} bathymetry: file found, skipping download')
        else:
            logging.info(f"CHS bathymetry: downloading {imgnum}/{len(imgs)} "
                          "from CHS NONNA-100...")
            assert(len(img['rasterIds']) == 1)
            url3 = f"{source}ImageServer/file?id={img['id'][0:]}&rasterId={img['rasterIds'][0]}"
            tiff = requests.get(url3)
            assert(tiff.status_code == 200)
            with open(fpath, "wb") as f: f.write(tiff.content)
            imgnum += 1

    logging.info(f'CHS bathymetry: processing {len(filepaths)} '
                 f'file{"s" if len(filepaths)!=1 else ""}')

    # read downloaded files and process them for DB insertion
    for filepath in filepaths:
        # open image and interpret pixels as elevation
        im = Image.open(filepath)
        nan = float(im.tag[42113][0])
        val = np.ndarray((im.size[0], im.size[1]))
        for yi in range(im.size[1]):
            val[:,yi] = np.array(list(map(im.getpixel, zip(
                    [yi for xi in range(im.size[0])], 
                    range(im.size[1])))))
        mask = np.flip(val == nan, axis=0)

        # generate latlon arrays
        file_south, file_west = parse_sw_corner(filepath)
        dlat = 0.001
        if file_south < 68:
            dlon = 0.001
        elif file_south >=68 and file_south < 80:
            dlon = 0.002
        elif file_south >= 80:
            dlon = 0.004
        file_xmax = im.size[0] * dlon + file_west
        file_ymax = im.size[1] * dlat + file_south
        file_lon = np.linspace(start=file_west,  stop=file_xmax, num=im.size[0])
        file_lat = np.linspace(start=file_south, stop=file_ymax, num=im.size[1])

        # select non-masked entries, remove missing, build grid
        z1 = np.flip(val, axis=0)
        x1, y1 = np.meshgrid(file_lon, file_lat)
        x2, y2, z2 = x1[~mask], y1[~mask], np.abs(z1[~mask])
        source = ['chs' for z in z2]
        grid = list(map(tuple, np.vstack((z2, y2, x2, source)).T))

        # insert into db
        n1 = db.execute(f"SELECT COUNT(*) FROM {chs_table}").fetchall()[0][0]
        db.executemany(f"INSERT OR IGNORE INTO {chs_table} VALUES (?,?,?,?)", grid)
        n2 = db.execute(f"SELECT COUNT(*) FROM {chs_table}").fetchall()[0][0]
        db.execute("COMMIT")
        conn.commit()
        logging.info(f"CHS {filepath.split('/')[-1]} bathymetry in region "
              f"{fmt_coords(dict(south=south,west=west,north=north,east=east))}. "
              f"processed and inserted {n2-n1} rows. "
              f"{len(z1[~mask]) - len(grid)} null values removed, "
              f"{len(grid) - (n2-n1)} duplicate rows ignored")

    return True


def load_chs(south, north, west, east):
    """ load bathymetric data from the database

        args:
            south, north:
                y-grid coordinate boundaries (float)
            west, east:
                x-grid coordinate boundaries (float)

        return:
            bathy:
                bathymetric values within query range
            lat:
                y-grid coordinate values
            lon:
                x-grid coordinate values
    """
    # check for missing data
    qryargs = dict(
            south=south, west=west,
            north=north, east=east, 
            start=datetime.now(), end=datetime.now())
    kadlu.geospatial.data_sources.fetch_handler.fetch_handler(
            'bathy', 'chs', parallel=False, **qryargs)

    # load the data
    db.execute(' AND '.join([f"SELECT * FROM {chs_table} WHERE lat >= ?",
                                                              "lat <= ?",
                                                              "lon >= ?",
                                                              "lon <= ?"]),
               tuple(map(str, [south, north, west, east])))
    
    rowdata = np.array(db.fetchall(), dtype=object).T
    #assert len(rowdata) == 4, "no data found for query range"
    if len(rowdata) == 0:
        logging.warning('CHS bathymetry: no data found, returning empty arrays')
        return np.array([[],[],[],[]])

    bathy, lat, lon, source = rowdata
    return np.array((bathy, lat, lon)).astype(float)


class Chs():
    """ collection of module functions for fetching and loading """

    def fetch_bathymetry(self, **kwargs):
        # trim query indexing entropy and check for fetched data
        for k in ('start', 'lock', 'end', 'top', 'bottom'):
            if k in kwargs.keys(): del kwargs[k]
        if serialized(kwargs, 'fetch_chs_bathy'): return False

        # if new data was fetched, index the query hash
        if (fetch_chs(south=kwargs['south'], north=kwargs['north'], 
                west=kwargs['west'], east=kwargs['east'], band_id=1)):
            insert_hash(kwargs, 'fetch_chs_bathy')
        return True

    def load_bathymetry(self, **kwargs):
        return load_chs(south=kwargs['south'], north=kwargs['north'], 
              west=kwargs['west'], east=kwargs['east'])

    def __str__(self):
        info = "\n".join(["Non-Navigational 100m (NONNA-100) bathymetry dataset",
            "from the Canadian Hydrographic Datastore",
            "\thttps://open.canada.ca/data/en/dataset/d3881c4c-650d-4070-bf9b-1e00aabf0a1d"])
        args = "(south, north, west, east)"
        return str_def(self, info, args)





""" Sound speed module within the kadlu package
"""
import gsw
import numpy as np
from kadlu.utils import interp_grid_1d, deg2rad
from kadlu.geospatial.interpolation import Interpolator2D, Interpolator3D, Uniform3D, DepthInterpolator3D


def sound_speed_teos10(lats, lons, z, t, SP):
    """ Compute sound speed from temperature, salinity, and depth 
        for a given latitude and longitude using the Thermodynamic 
        Equation of Seawater 2010.

        https://teos-10.github.io/GSW-Python/

        Args:
            lats: numpy array
                Latitudes (-90 to 90 degrees)
            lons: numpy array
                Longitudes (-180 to 180 degrees)
            z: numpy array
                Depths (meters)
            t: numpy array
                In-situ temperature (Celsius)
            SP: numpy array
                Practical Salinity (psu)

        Returns:
            c: numpy array
                Sound speed (m/s) 
    """
    p = gsw.p_from_z(z=-z, lat=lats)  # sea pressure (gsw uses negative z below sea surface)
    SA = gsw.SA_from_SP(SP, p, lons, lats)  # absolute salinity
    CT = gsw.CT_from_t(SA, t, p)  # conservative temperature
    c = gsw.density.sound_speed(SA=SA, CT=CT, p=p)
    return c

class SoundSpeed():
    """ Class for handling computation and interpolation of sound speed. 

        The sound speed can be specified via the argument ssp (sound speed 
        profile) or computed from the ocean variables (temperature, salinity).

        ssp can be either a single value, in which case the sound speed is 
        the same everywhere, or a tuple (c,z) where c is an array of sound 
        speed values and z is an array of depths.

        The interp and interp_xyz method may be used to obtain the interpolated 
        sound speed at any set of coordinates.

        Args:
            ocean: instance of :class:`kadlu.geospatial.ocean.Ocean`
                Ocean variables
            ssp: float or tuple
                Sound speed profile. May be specified either as a float, 
                in which case the sound speed is the same everywhere, or as 
                a tuple (c,z) where c is an array of sound speeds and z is 
                an array of depth values.
            num_depths: int
                Number of depth values for the interpolation grid. The default value is 50.
            rel_err: float
                Maximum deviation of the interpolation, expressed as a ratio of the 
                range of sound-speed values. The default value is 0.001.
    """
    def __init__(self, ocean=None, ssp=None, num_depths=50, rel_err=1E-3):

        assert ocean is not None or ssp is not None, "ocean or ssp must be specified"

        if ssp is not None:
            if isinstance(ssp, tuple): self._interp = DepthInterpolator3D(values=ssp[0], depths=ssp[1])
            else: self._interp = Uniform3D(values=ssp)

        else:
            lat_res, lon_res = self._lat_lon_res(ocean, default_res=1.0) #default resolution is 1 degree, approx 100 km

            # geographic boundaries
            S,N,W,E = ocean.boundaries['south'], ocean.boundaries['north'], ocean.boundaries['west'], ocean.boundaries['east']

            # lat and lon coordinates
            num_lats = max(3, int(np.ceil((N - S) / lat_res)) + 1)
            lats = np.linspace(S, N, num=num_lats)
            num_lons = max(3, int(np.ceil((E - W) / lon_res)) + 1)
            lons = np.linspace(W, E, num=num_lons)

            # compute depth coordinates
            depths = self._depth_coordinates(ocean, lats, lons, num_depths=num_depths, rel_err=rel_err)

            # interpolate temperature and salinity on lat,lon,depth grid
            t = ocean.temp(lat=lats, lon=lons, depth=depths, grid=True)
            s = ocean.salinity(lat=lats, lon=lons, depth=depths, grid=True)

            # compute sound speed
            grid_shape = t.shape
            la, lo, de = np.meshgrid(lats, lons, depths)
            la = la.flatten()
            lo = lo.flatten()
            de = de.flatten()
            t = t.flatten()
            s = s.flatten()
            c = sound_speed_teos10(lats=la, lons=lo, z=de, t=t, SP=s)
            c = np.reshape(c, newshape=grid_shape)

            # create interpolator
            self._interp = Interpolator3D(values=c, lats=lats, lons=lons,\
                    depths=depths, origin=ocean.origin, method='linear')

    def _lat_lon_res(self, ocean, default_res):
        """ Determine lat,lon resolutions for interpolation grid

            Args:
                ocean: instance of :class:`kadlu.geospatial.ocean.Ocean`
                    Ocean variables
                default_res: float
                    Default resolution in degrees

            Returns:
                lat_res, lon_res: float,float
                    Resolutions in degrees.
        """
        temp_nodes = ocean.interps['temp'].get_nodes()
        temp_lat, temp_lon = temp_nodes[1], temp_nodes[1]
        salinity_nodes = ocean.interps['salinity'].get_nodes()
        salinity_lat, salinity_lon = salinity_nodes[1], salinity_nodes[1]
        temp_lat_res = default_res if temp_lat is None else np.abs(temp_lat[1] - temp_lat[0])
        temp_lon_res = default_res if temp_lon is None else np.abs(temp_lon[1] - temp_lon[0])
        salinity_lat_res = default_res if salinity_lat is None else np.abs(salinity_lat[1] - salinity_lat[0])
        salinity_lon_res = default_res if salinity_lon is None else np.abs(salinity_lon[1] - salinity_lon[0])
        lat_res = min(temp_lat_res, salinity_lat_res)
        lon_res = min(temp_lon_res, salinity_lon_res)
        return lat_res, lon_res

    def _depth_coordinates(self, ocean, lats, lons, num_depths, rel_err):
        """ Compute depth coordinates for lat,lon,depth interpolation grid.

            Args:
                ocean: instance of :class:`kadlu.geospatial.ocean`
                    Ocean variables
                lats,lons: numpy.array, numpy.array
                    Latitude and longitude coordinates
                num_depths: int
                    Number of depth values for the interpolation grid. The default value is 50.
                rel_err: float
                    Maximum deviation of the interpolation, expressed as a ratio of the 
                    range of sound-speed values. The default value is 0.001.

            Returns:
                depths: numpy.array
                    Depth coordinates
        """
        seafloor_depth = ocean.bathy(lat=lats, lon=lons, grid=True)

        # find deepest point
        deepest_point = np.unravel_index(np.argmax(seafloor_depth), seafloor_depth.shape)

        # depth and lat,lon coordinates at deepest point
        max_depth = seafloor_depth[deepest_point]
        lat = lats[deepest_point[0]]
        lon = lons[deepest_point[1]]

        # compute temperature, salinity and sound speed for every 1 meter
        z = np.arange(0, int(np.ceil(max_depth))+1)
        t = ocean.temp(lat=lat, lon=lon, depth=z, grid=True)
        s = ocean.salinity(lat=lat, lon=lon, depth=z, grid=True)        
        c = sound_speed_teos10(lats=lat, lons=lon, z=z, t=t, SP=s)

        # determine grid
        indices, _ = interp_grid_1d(y=c, x=z, num_pts=num_depths, rel_err=rel_err)
        depths = z[indices]

        return depths

    def interp(self, lat, lon, z, grid=False):
        """ Interpolate sound speed in spherical coordinates (lat-lon).

            lat,lot,z can be floats or arrays.

            If grid is set to False, the interpolation will be evaluated at 
            the coordinates (lat_i, lon_i, z_i), where lat=(lat_1,...,lat_N), 
            lon=(lon_1,...,lon_N) and z=(z_,...,z_K). Note that in this case, lat and 
            lon must have the same length.

            If grid is set to True, the interpolation will be evaluated at 
            all combinations (lat_i, lon_j, z_k), where lat=(lat_1,...,lat_N), 
            lon=(lon_1,...,lon_M) and z=(z_1,...,z_K). Note that in this case, the lengths 
            of lat and lon do not have to be the same.

            Args: 
                lat: float or array
                    Latitude of the positions(s) where the bathymetry is to be evaluated
                lon: float or array
                    Longitude of the positions(s) where the bathymetry is to be evaluated
                z: float or array
                    Depth of the positions(s) where the interpolation is to be evaluated
                grid: bool
                    Specify how to combine elements of lat,lon,z.

            Returns:
                : numpy.array
                    Interpolated sound speed values
        """
        return self._interp.interp(lat=lat, lon=lon, z=z, grid=grid)

    def interp_xy(self, x, y, z, grid=False):
        """ Interpolate sound speed in planar (x-y) geometry.

            x,y,z can be floats or arrays.

            If grid is set to False, the interpolation will be evaluated at 
            the positions (x_i, y_i, z_i), where x=(x_1,...,x_N),  
            y=(y_1,...,y_N), and z=(z_1,...,z_N). Note that in this case, 
            x,y,z must have the same length.

            If grid is set to True, the interpolation will be evaluated at 
            all combinations (x_i, y_j, z_k), where x=(x_1,...,x_N), 
            y=(y_1,...,y_M), and z=(z_1,...,z_K). Note that in this case, the 
            lengths of x,y,z do not have to be the same.

            Args: 
                x: float or array
                    x-coordinates
                y: float or array
                    y-coordinates
                z: float or array
                    depth(s)
                grid: bool
                   Specify how to combine elements of x,y,z.

            Returns:
                : numpy.array
                    Interpolated sound speed values
        """
        return self._interp.interp_xy(x=x, y=y, z=z, grid=grid)


